import re
from typing import Union, List, TextIO

from .consoles import SSHConsole, LocalConsole
from .pci_parser import PCIParser, PCIDevice


class DoesNotExist(Exception):
    def __init__(self, **kwargs):
        self.arguments = kwargs

    def __str__(self):
        return 'Unable to find PCI Device matching: {query}'.format(query=str(self.arguments))


# noinspection PyBroadException
class ScannerPCI(object):
    def __init__(self, ip: str, username: str = None, password: str = None, port: int = 22, logfile: TextIO = None):
        self.ip = ip
        self.username = username
        self.password = password
        self.port = port
        self.logfile = logfile
        self._parser: Union[PCIParser, None] = None
        self._console: Union[SSHConsole, LocalConsole, None] = None

    def _get_console(self) -> Union[SSHConsole, LocalConsole]:
        if self._console is None:
            if self.ip == '127.0.0.1' or self.ip == 'localhost':
                self._console = LocalConsole(
                    password=self.password,
                    logfile=self.logfile
                )
            else:
                self._console = SSHConsole(
                    ip=self.ip,
                    username=self.username,
                    password=self.password,
                    port=self.port,
                    logfile=self.logfile
                )
        return self._console

    def _get_pci_addresses(self) -> List[str]:
        """
        Get all PCI addresses from target machine device
        :return: List of PCI addresses
        """
        result: List[str] = []
        try:
            command = "lspci -D | grep -o -e '[0-9A-Za-z]\{4\}:[0-9A-Za-z]\{2\}:[0-9A-Za-z]\{2\}.[0-9A-Za-z]\{1\}' | sed $'s/[^[:print:]]//g'"
            reply = self._get_console().run_command(command=command, sudo=True)
            for line in reply:
                if bool(re.match(r"""[0-9a-fA-F]{4}:[0-9a-fA-F]{2}:[0-9a-fA-F]{2}.[0-9a-fA-F]""", line)):
                    result.append(line)
        except Exception as error:
            print(error)
        finally:
            return result

    def _scan_single_pci_device(self, address: str) -> Union[List[str], None]:
        """
        Scan PCI device
        :param address: Address target to scan
        :return: PCI device data or Nothing
        """
        result = None
        try:
            command = 'lspci -s {address} -D -vv'.format(address=address)
            result = self._get_console().run_command(command=command, sudo=True)
        except Exception as error:
            print(error)
        finally:
            return result

    def _get_pci(self, force_rescan: bool = False) -> List[PCIDevice]:
        if self._parser is None or force_rescan is True:
            self._parser = PCIParser(data=[self._scan_single_pci_device(address=address) for address in self._get_pci_addresses()])
        return self._parser.devices

    def pci_rescan(self) -> None:
        self._get_console().run_command(command="sh -c 'echo 1 > /sys/bus/pci/rescan'", sudo=True)

    def select(self, force_rescan: bool = False, **kwargs) -> List[PCIDevice]:
        if len(kwargs.items()) != 0:
            return [d for d in self._get_pci(force_rescan=force_rescan) if d.match(**kwargs)]
        else:
            return self._get_pci(force_rescan=force_rescan)

    def get(self, force_rescan: bool = False, **kwargs):
        result = self.select(force_rescan=force_rescan, **kwargs)
        if len(result) != 0:
            return result[0]
        else:
            raise DoesNotExist(**kwargs)

    def get_connected(self, parent: PCIDevice, force_rescan: bool = False) -> List[PCIDevice]:
        if parent.is_host_bridge:
            return [d for d in self._get_pci(force_rescan=force_rescan) if d.addr_dom == parent.addr_dom and d.addr_bus == parent.addr_bus and d != parent]
        elif parent.is_root_port or parent.is_downstream:
            return [d for d in self._get_pci(force_rescan=force_rescan) if d.addr_dom == parent.addr_dom and d.addr_bus == parent.pci_bus.secondary]
        elif parent.is_upstream:
            return [d for d in self._get_pci(force_rescan=force_rescan) if d.is_downstream and d.pci_bus.primary == parent.pci_bus.secondary]
        else:
            return []



import time
from typing import List, Union, TextIO

import paramiko

from .console import Console, CommandFailed


class SSHConnectionError(Exception):
    """
    Connection error occurs.
    """

    def __init__(self, message: str) -> None:
        self.message = message

    def __str__(self) -> str:
        return self.message


# noinspection PyBroadException
class SSHConsole(Console):
    def __init__(self, ip: str, username: str, password: str, port: int = 22, logfile: TextIO = None):
        self.ip = ip
        self.username = username
        self.password = password
        self.port = port
        self.logfile = logfile
        self._paramiko_client: Union[paramiko.SSHClient, None] = None

    def _new_client(self) -> paramiko.SSHClient:
        """
        Creates new instance of  connection, set its controls and return it.
        :return: Instance of paramiko client.
        """
        ssh = paramiko.SSHClient()
        ssh.set_missing_host_key_policy(paramiko.AutoAddPolicy())
        self._paramiko_client = ssh
        return self._paramiko_client

    def is_connected(self) -> bool:
        """
        :return: Current connection state
        """
        state = False
        if self._paramiko_client is not None:
            try:
                state = self._paramiko_client.get_transport().is_active()
            except Exception:
                pass
        return state

    def connect(self) -> None:
        """
        This procedure initiate connection. Creates paramiko client if needed and connects it to target machine.
        :return: Nothing.
        """
        self.terminate()
        self._new_client().connect(
            hostname=self.ip,
            port=self.port,
            username=self.username,
            password=self.password,
            look_for_keys=False,
            allow_agent=False)

    def terminate(self):
        """
        Terminate current paramiko client session.
        :return: Nothing.
        """
        if self.is_connected():
            self._paramiko_client.close()

    def _get_console(self, attempts: int = 60, retry_timeout: Union[int, float] = 2) -> paramiko.SSHClient:
        """
        Get or create and get ssh connection.
        :param retry_timeout: Timeout between connection attempts.
        :param attempts: How many attempts of reconnection should be before Exception.
        :return: Instance of paramiko client.
        """
        counter = 0
        while not self.is_connected():
            if counter > 0:
                time.sleep(retry_timeout)
            self.connect()
            counter += 1
            if counter > attempts:
                raise SSHConnectionError("Cannot login via SSH")
        return self._paramiko_client

    def run_command(self, command: str, timeout: int = 60, sudo: bool = True, suppress_logs: bool = False) -> List[str]:
        """
        Executes command on target machine and return output as list of string.
        After execution procedure will check command exit code and will raise CommandFailed if command failed.
        :param sudo: Run command with sudo privileges in case if user isn't root.
        :param command: String with command to execute.
        :param timeout: Timeout of command execution. In case if command will not return exit code before timeout expired exception will be raised.
        :param suppress_logs: boolean flag  if set as True will suppress log output in logfile.
        :return: List of output string.
        """
        if self.username != 'root' and sudo is True:
            command = "sudo -S -p '' " + command

        if self.logfile is not None and suppress_logs is False:
            self.logfile.write(('%s@%s:~$ %s' % (self.username, self.ip, command)))

        stdin, stdout, stderr = self._get_console().exec_command(command=command, timeout=timeout)

        if self.username != 'root' and sudo is True:
            time.sleep(0.1)
            stdin.write(self.password + "\n")
            stdin.flush()
        exitcode: int = stdout.channel.recv_exit_status()
        output = [line.decode('utf-8') for line in stdout.read().splitlines()]

        if self.logfile is not None and suppress_logs is False:
            for line in output:
                self.logfile.write(line)

            self.logfile.write(('%s@%s:~$ %s' % (self.username, self.ip, 'echo $?')))
            self.logfile.write(('%s@%s:~$ %s' % (self.username, self.ip, exitcode)))

        if exitcode != 0:
            raise CommandFailed(command, output, exitcode)
        return output

    def run_command_ignore_fail(self, command: str, timeout: int = 60, sudo: bool = True, suppress_logs: bool = False) -> List[str]:
        """
        Executes command on target machine and return output as list of string.
        :param sudo: Run command with sudo privileges in case if user isn't root.
        :param command: String with command to execute.
        :param timeout: Timeout of command execution. In case if command will not return exit code before timeout expired exception will be raised.
        :param suppress_logs: boolean flag  if set as True will suppress log output in logfile.
        :return: List of output string.
        """
        try:
            output: List[str] = self.run_command(command, timeout, sudo, suppress_logs)
        except CommandFailed as error:
            output = error.output
        return output



from abc import *
from typing import List


class CommandFailed(Exception):
    """
    Running a command failed with non-zero exit code.
    """

    def __init__(self, command: str, output: List[str], exitcode: int) -> None:
        self.command = command
        self.output = output
        self.exitcode = exitcode

    def __str__(self) -> str:
        return "Command '{command}' exited with {code}. Output: {output}".format(
            command=self.command, code=self.exitcode, output=self.output)


class Console(ABC):
    @abstractmethod
    def run_command(self, command: str, sudo: bool = True) -> List[str]:
        pass

    @abstractmethod
    def run_command_ignore_fail(self, command: str, sudo: bool = True) -> List[str]:
        pass



import os
from typing import List, TextIO

from .console import Console, CommandFailed


class LocalConsole(Console):
    def __init__(self, password: str = None, logfile: TextIO = None):
        self.password = password
        self.logfile = logfile

    def run_command(self, command: str, sudo: bool = True, suppress_logs: bool = False) -> List[str]:
        """
        Executes command on Local Machine and return output as list of string.
        After execution procedure will check command exit code and will raise CommandFailed if command failed.
        :param command:  String with command to execute.
        :param sudo: Run command with sudo privileges in case if user isn't root.
        :param suppress_logs: boolean flag  if set as True will suppress log output in logfile.
        :return: List of output string.
        """
        if self.logfile is not None and suppress_logs is False:
            self.logfile.write(command)

        if sudo is True:
            command = 'echo %s|sudo -S %s' % (self.password, command)

        process = os.popen(command)
        output = process.read().splitlines()
        exitcode = process.close()

        if self.logfile is not None and suppress_logs is False:
            for line in output:
                self.logfile.write(line)

        if exitcode is not None:
            raise CommandFailed(command, output, (exitcode >> 8))
        else:
            return output

    def run_command_ignore_fail(self, command: str, sudo: bool = True, suppress_logs: bool = False) -> List[str]:
        """
        Executes command on Local Machine, ignores error exit codes and return output as list of string.
        :param command:  String with command to execute.
        :param sudo: Run command with sudo privileges in case if user isn't root.
        :param suppress_logs: boolean flag  if set as True will suppress log output in logfile.
        :return: List of output string.
        """
        try:
            output: List[str] = self.run_command(command=command, sudo=sudo, suppress_logs=suppress_logs)
        except CommandFailed as error:
            output = error.output
        return output
